/*------------------------------------------------------------------------------
* ss2.c : superstar II receiver dependent functions
*
*          Copyright (C) 2007,2008 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2008/07/14 00:05:05 $
* history : 2008/05/18 1.0 new
*           2008/06/16 1.2 separate common functions to rcvcmn.c
*-----------------------------------------------------------------------------*/
#include "rtklib.h"
#include "ss2.h"

#define NTOBS       3
#define SS2SOH      0x01        /* ss2 start of header */
#define SS2LLH      20          /* ss2 message ID#20 navigation data (user) */
#define SS2ECEF     21          /* ss2 message ID#21 navigation data (ecef) */
#define SS2EPH      22          /* ss2 message ID#22 ephemeris data */
#define SS2RAW      23          /* ss2 message ID#23 measurement block */
#define SS2SBAS     67          /* ss2 message ID#67 sbas data */
#define U1(p)       (*((unsigned char *)(p)))
#define U2(p)       (*((unsigned short *)(p)))
#define U4(p)       (*((unsigned int *)(p)))
#define R8(p)       (*((double *)(p)))

static const char rcsid[]="$Id: ss2.c,v 1.2 2008/07/14 00:05:05 TTAKA Exp $";

/* output rinex obs types */
static char tobs[NTOBS][3]={"C1","L1","S1"};

/* application defined functions ---------------------------------------------*/
extern int showmsg(char *format,...);

/* extract unsigned bits from message ----------------------------------------*/
static unsigned int getbitu(const unsigned char *buff, int pos, int len)
{
    unsigned int bit=0;
    int i;
    
    for (i=pos;i<pos+len;i++) bit=(bit<<1)+((buff[i/8]>>(7-i%8))&1U);
    return bit;
}
/* check checksum ------------------------------------------------------------*/
static int chksum(const unsigned char *buff, int len)
{
    int i,stat;
    unsigned short sum=0;
    
    for (i=0;i<len-2;i++) sum+=buff[i];
    if (!(stat=((sum>>8)==buff[len-1])&&((sum&0xFF)==buff[len-2])))
        fprintf(stderr,"chksum : msg=%2d sum=%02X%02X %02X%02X\n",
                buff[1],sum>>8,sum&0xFF,buff[len-1],buff[len-2]);
    return stat;
}
/* adjust time ---------------------------------------------------------------*/
static gtime_t adjtime(gtime_t time, double tow)
{
    double t;
    int week;
    if (time.time==0) return time;
    t=time2gpst(time,&week);
    if (fabs(t-tow)>302400.0) week+=t<tow?-1:1;
    return gpst2time(week,tow);
}
/* decode id#20 navigation data (user) ---------------------------------------*/
static gtime_t decodess2llh(const unsigned char *buff, int len)
{
	double ep[6];
    ep[3]=U1(buff); ep[4]=U1(buff+1); ep[5]=R8(buff+2);
	ep[2]=U1(buff+10); ep[1]=U1(buff+11); ep[0]=U2(buff+12);
	return utc2gpst(epoch2time(ep));
}
/* decode id#21 navigation data (ecef) ---------------------------------------*/
static gtime_t decodess2ecef(const unsigned char *buff, int len)
{
	return gpst2time(U2(buff+4),R8(buff));
}
/* decode id#23 measurement block --------------------------------------------*/
static int decodess2meas(raw_t *raw, const unsigned char *buff, int len,
                         int opt, obsd_t *obs)
{
    const double freqif=1.405396825E6,tslew=1.75E-7;
    double tow,slew,code,icp,d;
    int i,n,sat,nobs,val;
    const unsigned char *q;
    unsigned int sc;
    
    nobs=U1(buff+2);
    slew=U1(buff)*tslew;
    tow=floor(R8(buff+3)*1000.0+0.5)/1000.0; /* rounded by 1ms */
    raw->icpc+=4.5803-freqif*slew-FREQ1*(slew-1E-6); /* phase correction */
    raw->time=adjtime(raw->time,tow);
    
    val=screent(raw->time,raw->ts,raw->te,raw->tint);
    
    for (i=n=0,q=buff+11;i<nobs&&n<MAXOBS;i++,q+=11) {
        sat=(q[0]&0x1F)+1;
        obs[n].time=raw->time;
        obs[n].sat=sat;
        code=(tow-floor(tow))-(double)(U4(q+2))/2095104000.0;
        obs[n].P[0]=CLIGHT*(code+(code<0.0?1.0:0.0));
        icp=(double)(U4(q+6)>>2)/1024.0+raw->off[sat-1]; /* unwrap */
        if (fabs(icp-raw->icpp[sat-1])>524288.0) {
            d=icp>raw->icpp[sat-1]?-1048576.0:1048576.0;
            raw->off[sat-1]+=d; icp+=d;
        }
        raw->icpp[sat-1]=icp;
        obs[n].L[0]=icp+raw->icpc;
        obs[n].D[0]=0.0;
        obs[n].SNR[0]=(unsigned char)(floor(U1(q+1)*0.25+0.5));
        sc=U1(q+10);
        obs[n].LLI[0]=(int)((unsigned char)sc-(unsigned char)raw->lockc[sat-1][0])>0;
        obs[n].LLI[0]|=U1(q+6)&1?2:0;
        if (val) raw->lockc[sat-1][0]=sc;
        n++;
    }
    return val?n:0;
}
/* decode id#22 ephemeris data ------------------------------------------------*/
static int decodess2eph(raw_t *raw, const unsigned char *buff, int len,
                        eph_t *ephs)
{
    eph_t eph={0};
    double tow,ion[8],utc[4];
    unsigned int words[10]={0};
    int i,j,sat,week,leaps,iode1=-1,iode2=-1;
    
    sat=(U1(buff)&0x1F)+1;
    eph.sat=(unsigned char)sat;
    tow=time2gpst(raw->time,&week);
    
    for (i=0;i<3;i++) {
        words[1]=(((int)tow/6)<<7)+((unsigned char)(i+1)<<2);
        for (j=2;j<10;j++) words[j]=getbitu(buff+1+i*24,(j-2)*24,24);
        if (decodefrm(words,&eph,ion,utc,&leaps)!=i+1) return 0;
        if      (i==0) iode1=eph.iodc&0xFF;
        else if (i==1) iode2=eph.iode;
        else if (eph.iode!=iode1||eph.iode!=iode2) return 0;
    }
    eph.ttr=adjtime(raw->time,time2gpst(eph.ttr,NULL));
    eph.toe=adjtime(raw->time,eph.toes);   
    ephs[sat-1]=eph;
    return sat;
}
/* decode id#67 sbas data ----------------------------------------------------*/
static int decodess2sbas(const unsigned char *buff, int len, sbsmsg_t *sbsmsg)
{
    gtime_t time;
    int i,sat;
    
    sat=U4(buff+12);
    if (len!=48||sat<MINPRNSBS||MAXPRNSBS<sat) return 0;
    sbsmsg->week=U4(buff);
    sbsmsg->tow=(int)R8(buff+4);
    time=gpst2time(sbsmsg->week,sbsmsg->tow);
    sbsmsg->sat=sat;
    for (i=0;i<29;i++) sbsmsg->msg[i]=buff[16+i];
    return sat;
}
/* decode superstar II message -------------------------------------------------
* decode and input a novatel message. support ID#22 ephemeris data, ID#23
* measurement block, ID#67 sbas data
* args   : raw_t *raw    IO     raw decode control struct
*          unsigned char *buff I message buffer
*          int   len     I      buffer length (bytes)
*          int   opt     I      decode option
*          obs_t *obs    IO     observation data (NULL: no output)
*          nav_t *nav    IO     navigation message (NULL: no output)
*          sbsmsg_t *sbsmsg IO  sbas message (NULL: no output)
*          int   *sat    O      satellite prn number for ephemeris or sbas msg
* return : input status (0:no valid data, 1:obs data, 2:ephemeris,
*                        3:sbas message, 9:ion/utc parameters)
*-----------------------------------------------------------------------------*/
extern int decodess2(raw_t *raw, const unsigned char *buff, int len, int opt,
                     obs_t *obs, nav_t *nav, sbsmsg_t *sbsmsg, int *sat)
{
    if (!chksum(buff,len)) return -1;
    
    switch (buff[1]) {
    case SS2LLH: /* id#20 */
        raw->time=decodess2llh(buff+4,buff[3]);
        break;
    case SS2ECEF: /* id#21 */
        raw->time=decodess2ecef(buff+4,buff[3]);
        break;
    case SS2RAW: /* id#23 */
        if (raw->time.time==0) break;
        if (obs&&(obs->n=decodess2meas(raw,buff+4,buff[3],opt,obs->data))>0) {
            return 1;
        }
        break;
    case SS2EPH: /* id#22 */
        if (raw->time.time==0) break;
        if (nav&&(*sat=decodess2eph(raw,buff+4,buff[3],nav->eph))>0) {
            return screent(nav->eph[*sat-1].ttr,raw->ts,raw->te,0.0)?2:0;
        }
        break;
    case SS2SBAS: /* id#67 */
        if (sbsmsg&&(*sat=decodess2sbas(buff+4,buff[3],sbsmsg))>0) {
            return screent(gpst2time(sbsmsg->week,sbsmsg->tow),raw->ts,raw->te,0.0)?3:0;
        }
        break;
    }
    return 0;
}
/* convert superstar II log file -----------------------------------------------
* convert superstart II log file to rinex obs/nav and sbas message file.
* input log file shall contain ID#22, ID#23, ID#67
* args   : FILE  *fp     I      input log file pointer
*          FILE  *ofp    I      output rinex obs file pointer (NULL:no output)
*          FILE  *nfp    I      output rinex nav file pointer (NULL:no output)
*          FILE  *sfp    I      output sbas message file pointer (NULL:no output)
*          gtime_t ts    I      time start
*          gtime_t te    I      time end
*          double tint   I      time interval
*          rnxopt_t rnxopt I    rinex options
* return : none
*-----------------------------------------------------------------------------*/
extern void convss2(FILE *fp, FILE *ofp, FILE *nfp, FILE *sfp, gtime_t ts,
                    gtime_t te, double tint, const rnxopt_t *rnxopt, int opt)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obs={0};
    nav_t nav={0};
    sbsmsg_t sbsmsg;
    raw_t raw={{0}};
    gtime_t time={0},tstart={0};
    int c,sat,len,type,nmsg[4]={0};
    char s[128]="",s1[64],s2[64];
    unsigned char buff[512];
    
    obs.data=data; nav.eph=eph;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        if (c!=SS2SOH) continue;
        buff[0]=(unsigned char)c;
        if (fread(buff+1,1,3,fp)<3) break;
        if ((buff[1]^buff[2])!=0xFF) continue;
        len=buff[3];
        if (len+6>sizeof(buff)) {nmsg[3]++; continue;}
        if (fread(buff+4,len+2,1,fp)<1) break;
        
        type=decodess2(&raw,buff,len+6,opt,&obs,&nav,&sbsmsg,&sat);
        
        if (type==1&&ofp) {
            if (nmsg[0]==0) {
                outrnxobsh(ofp,rnxopt,tobs,NTOBS,obs.data[0].time);
            }
            outrnxobsb(ofp,obs.data,obs.n,0,tobs,NTOBS);
            nmsg[0]++;
            time=obs.data[0].time;
        }
        else if (type==2&&nfp) {
            if (nmsg[1]==0) {
                outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
            }
            outrnxnavb(nfp,&nav.eph[sat-1]);
            nmsg[1]++;
            if (!ofp) time=nav.eph[sat-1].ttr;
        }
        else if (type==3&&sfp) {
            sbsoutmsg(sfp,&sbsmsg);
            nmsg[2]++;
            if (!ofp&&!nfp) time=gpst2time(sbsmsg.week,sbsmsg.tow);
        }
        else if (type==-1) nmsg[3]++;
        if (te.time!=0&&timediff(obs.data[0].time,te)>DTTOL+1.0) break;
        if (time.time!=0) {
            time2str(time,s1,0); sprintf(s,"%s: ",s1);
            if (tstart.time==0) tstart=time;
        }
        if (showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],
                    nmsg[2],nmsg[3])) break;
    }
    if (nfp) {
        rewind(nfp);
        outrnxnavh(nfp,rnxopt,nav.ion,nav.utc,nav.leaps);
    }
    if (tstart.time!=0) {
        time2str(tstart,s1,0); time2str(time,s2,0);
        sprintf(s,"%s-%s: ",s1,s2+5);
    }
    showmsg("%sObs=%d Nav=%d Sbs=%d Err=%d",s,nmsg[0],nmsg[1],nmsg[2],nmsg[3]);
}
/* read super starr II log file ------------------------------------------------
* read superstar II log file. input log file shall contain ID#22, ID#23, ID#67
* args   : FILE  *fp     I      input file pointer
*          gtime_t ts,te I      time start/end
*          double tint   I      time interval
*          int   opt     I      read option)
*          obs_t *obs    IO     observation data    (NULL: no output)
*          nav_t *nav    IO     navigation messages (NULL: no output)
*          sbs_t *sbs    IO     sbas messages       (NULL: no output)
*          int   *nmax   I      max number of messages 
* return : none
*-----------------------------------------------------------------------------*/
extern void readss2(FILE *fp, gtime_t ts, gtime_t te, double tint, int opt,
                    obs_t *obs, nav_t *nav, sbs_t *sbs, int *nmax)
{
    obsd_t data[MAXOBS]={{{0}}};
    eph_t eph[32]={{0}};
    obs_t obsp={0};
    nav_t navp={0};
    sbsmsg_t sbsp;
    raw_t raw={{0}};
    int c,sat,len,type;
    unsigned char buff[512];
    
    obsp.data=data; navp.eph=eph;
    raw.ts=ts; raw.te=te; raw.tint=tint;
    
    while ((c=fgetc(fp))!=EOF) {
        if (c!=SS2SOH) continue;
        buff[0]=(unsigned char)c;
        if (fread(buff+1,1,3,fp)<3) break;
        if (buff[1]!=(buff[2]^0xFF)) continue;;
        len=buff[3];
        if (len+6>sizeof(buff)) continue;
        if (fread(buff+4,1,len+2,fp)<(size_t)len+2) break;
        
        type=decodess2(&raw,buff,len+6,opt,&obsp,&navp,&sbsp,&sat);
        
        if (type==1&&obs) {
            addobs(obs,&obsp,nmax);
        }
        else if (type==2&&nav) {
            addnav(nav,&navp,sat,nmax+1);
        }
        else if (type==3&&sbs) {
            addsbs(sbs,sbsp,nmax+2);
        }
        else if (type==9&&nav) {
            addionutc(nav,&navp);
        }
        if (te.time!=0&&timediff(obsp.data[0].time,te)>DTTOL) break;
    }
}
