/*------------------------------------------------------------------------------
* rcvlog.c : receiver log functions
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2008/06/18 1.0 new
*           2009/01/06 1.1 add function decodelog, decodeobs
*-----------------------------------------------------------------------------*/
#include "rtklib.h"
#include "rcv/novatel.h"
#include "rcv/ublox.h"
#include "rcv/ss2.h"
#include "rcv/crescent.h"
#include "rcv/garmin.h"

static const char rcsid[]="$Id: rcvlog.c,v 1.1 2008/07/17 21:48:06 ttaka Exp $";

/* extract bits from navigation data word (pos=1,2,...) ----------------------*/
#define BIT(word,pos,len) (((word)>>(25-(pos)-(len)))&~((~0U)<<(len)))

/* convert navigation data bits to real (len<=32) ----------------------------*/
static double toreal(unsigned int bits, int len, int sign, int fact)
{
    if (!sign) return bits*pow(2.0,fact);
    if (bits>>(len-1)&&len<32) bits|=(~0U)<<len; /* extend bits if minus */
    return (int)bits*pow(2.0,fact);
}
/* decode navigation data frame ------------------------------------------------
* decode gps navigation data frame.
* args   : unsigned int *words I navigation data frame(24bitx10) without parity
*          eph_t *eph    O      ephemeris message
*          double *ion   O      ionospheric parameters
*          double *utc   O      delta-utc parameters
*          int   *leaps  O      leap seconds (s)
* return : subframe id (1-3: ephemeris frame, 9:ion/utc parameters)
*-----------------------------------------------------------------------------*/
extern int decodefrm(const unsigned int *words, eph_t *eph, double *ion,
                     double *utc, int *leaps)
{
    double sqrtA;
    int week,tow,id;
    
    tracet(4,"decodefrm:\n");
    
    tow=BIT(words[1],1,17)*6;
    id=BIT(words[1],20,3);
    switch (id) {
    case 1: /* subframe 1 */
        week     =BIT(words[2],1,10)+OFFWEEK;
        eph->ttr =gpst2time(week,tow);
        eph->sva =BIT(words[2],13,4);
        eph->svh =BIT(words[2],17,6);
        eph->iodc=(BIT(words[2],23,2)<<8)+(BIT(words[7],1,8));
        eph->tgd =toreal(BIT(words[6],17,8),8,1,-31);
        eph->toc =gpst2time(week,toreal(BIT(words[7],9,16),16,0,4));
        if (timediff(eph->toc,eph->ttr)<-302400.0)
            eph->toc=timeadd(eph->toc,604800.0);
        eph->f2  =toreal(BIT(words[8],1,8),8,1,-55);
        eph->f1  =toreal(BIT(words[8],9,16),16,1,-43);
        eph->f0  =toreal(BIT(words[9],1,22),22,1,-31);
        break;
    case 2: /* subframe 2 */
        eph->iode=BIT(words[2],1,8);
        eph->crs =toreal(BIT(words[2],9,16),16,1,-5);
        eph->deln=toreal(BIT(words[3],1,16),16,1,-43)*SC2RAD;
        eph->M0  =toreal((BIT(words[3],17,8)<<24)+BIT(words[4],1,24),32,1,-31)*SC2RAD;
        eph->cuc =toreal(BIT(words[5],1,16),16,1,-29);
        eph->e   =toreal((BIT(words[5],17,8)<<24)+BIT(words[6],1,24),32,0,-33);
        eph->cus =toreal(BIT(words[7],1,16),16,1,-29);
        sqrtA    =toreal((BIT(words[7],17,8)<<24)+BIT(words[8],1,24),32,0,-19);
        eph->A   =sqrtA*sqrtA;
        eph->toes=toreal(BIT(words[9],1,16),16,0,4);
        eph->fit =BIT(words[9],17,1);
        break;
    case 3: /* subframe 3 */
        eph->cic =toreal(BIT(words[2],1,16),16,1,-29);
        eph->OMG0=toreal((BIT(words[2],17,8)<<24)+BIT(words[3],1,24),32,1,-31)*SC2RAD;
        eph->cis =toreal(BIT(words[4],1,16),16,1,-29);
        eph->i0  =toreal((BIT(words[4],17,8)<<24)+BIT(words[5],1,24),32,1,-31)*SC2RAD;
        eph->crc =toreal(BIT(words[6],1,16),16,1,-5);
        eph->omg =toreal((BIT(words[6],17,8)<<24)+BIT(words[7],1,24),32,1,-31)*SC2RAD;
        eph->OMGd=toreal(BIT(words[8],1,24),24,1,-43)*SC2RAD;
        eph->iode=BIT(words[9],1,8);
        eph->idot=toreal(BIT(words[9],9,14),14,1,-43)*SC2RAD;
        break;
    case 4: /* subframe 4 */
        if (BIT(words[2],3,6)==56) { /* page18 */
            ion[0]=toreal(BIT(words[2], 9,8),8,1,-30);
            ion[1]=toreal(BIT(words[2],17,8),8,1,-27);
            ion[2]=toreal(BIT(words[3], 1,8),8,1,-24);
            ion[3]=toreal(BIT(words[3], 9,8),8,1,-24);
            ion[4]=toreal(BIT(words[3],17,8),8,1,11);
            ion[5]=toreal(BIT(words[4], 1,8),8,1,14);
            ion[6]=toreal(BIT(words[4], 9,8),8,1,16);
            ion[7]=toreal(BIT(words[4],17,8),8,1,16);
            utc[0]=toreal((BIT(words[6],1,24)<<8)+BIT(words[7],1,8),32,1,-30);
            utc[1]=toreal(BIT(words[5],1,24),24,1,-50);
            utc[2]=toreal(BIT(words[7],9,8),8,0,12);
            utc[3]=BIT(words[7],17,8)+OFFWEEK;
            *leaps=(int)toreal(BIT(words[8],1,8),8,1,0);
            id=9;
        }
        break;
    }
    return id;
}
/* decode gps navigation data --------------------------------------------------
* decode gps navigation data frame and update navigation message.
* args   : raw_t  raw    IO     raw data control struct
*          int    sat    I      satellite PRN number
*          unsigned int *words I navigation data frame(24bitx10) without parity
*          nav    *nav   IO     navigation message
* return : decode status (0:no valid data, 2:ephemeris, 9:ion-utc)
*-----------------------------------------------------------------------------*/
extern int decodenav(raw_t *raw, int sat, const unsigned int *words, nav_t *nav)
{
    double ttr;
    int id,week;
    
    tracet(4,"decodenav: sat=%d\n",sat);
    
    if (sat<1||MAXPRNGPS<sat) return 0;
    
    raw->eph[sat-1].sat=sat;
    
    /* decode frame */
    id=decodefrm(words,raw->eph+sat-1,nav->ion,nav->utc,&nav->leaps);
    
    if (id==1) {
        raw->iode1[sat-1]=raw->eph[sat-1].iodc&0xFF;
    }
    else if (id==2) {
        raw->iode2[sat-1]=raw->eph[sat-1].iode;
    }
    else if (id==3&&raw->iode1[sat-1]==raw->eph[sat-1].iode&&
                    raw->iode2[sat-1]==raw->eph[sat-1].iode) {
        /* adjust reception time */
        ttr=time2gpst(raw->eph[sat-1].ttr,&week);
        if (ttr>302400.0&&raw->eph[sat-1].toes<302400.0) {
            raw->eph[sat-1].ttr=gpst2time(++week,ttr-604800.0);
        }
        /* adjust toe */
        raw->eph[sat-1].toe=gpst2time(week,raw->eph[sat-1].toes);
        
        /* copy ephemeris buffer to navigation messages */
        nav->eph[sat-1]=raw->eph[sat-1];
        return 2;
    }
    else if (id==9) return 9;
    
    return 0;
}
/* add observation data record -------------------------------------------------
* add observation data record to observation data
* args   : obs_t  *obs   IO     observation data
*          obsd_t *obsp  I      observation data record
*          int    *nmax  IO     max number of observation data records
*                               (if not enough size, expand size)
* return : none
*-----------------------------------------------------------------------------*/
extern void addobs(obs_t *obs, obs_t *obsp, int *nmax)
{
    int i;
    
    tracet(4,"addobs:\n");
    
    if (!obs) return;
    if (*nmax<=obs->n+obsp->n) {
        if (*nmax==0) *nmax=65532; else *nmax+=65532;
        if (!(obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*(*nmax)))) {
            fprintf(stderr,"memory allocation error\n");
            obs->n=0;
            return;
        }
    }
    for (i=0;i<obsp->n;i++) obs->data[obs->n++]=obsp->data[i];
}
/* add ephemeris records -------------------------------------------------------
* add ephemeris records to navigation messages
* args   : nav_t  *nav   IO     navigation messages
*          nav_t  *navp  I      ephemeris records
*          int     sat   I      satellite prn number
*          int    *nmax  IO     max number of ephemeris records
*                               (if not enough size, expand size)
* return : none
*-----------------------------------------------------------------------------*/
extern void addnav(nav_t *nav, nav_t *navp, int sat, int *nmax)
{
    tracet(4,"addnav:\n");
    
    if (!nav) return;
    if (*nmax<=nav->n) {
        if (*nmax==0) *nmax=1024; else *nmax+=1024;
        if (!(nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*(*nmax)))) {
            fprintf(stderr,"memory allocation error\n");
            nav->n=0;
            return;
        }
    }
    nav->eph[nav->n++]=navp->eph[sat-1];
}
/* add sbas message record -----------------------------------------------------
* add sbas message record to sbas messages
* args   : sbs_t  *sbs   IO     sbas messages
*          sbsmsg_t sbsp I      sbas message
*          int    *nmax  IO     max number of sbas message records
*                               (if not enough size, expand size)
* return : none
*-----------------------------------------------------------------------------*/
extern void addsbs(sbs_t *sbs, sbsmsg_t sbsp, int *nmax)
{
    tracet(4,"addsbs:\n");
    
    if (!sbs) return;
    if (*nmax<=sbs->n) {
        if (*nmax==0) *nmax=65532; else *nmax+=65532;
        if (!(sbs->msgs=(sbsmsg_t *)realloc(sbs->msgs,sizeof(sbsmsg_t)*(*nmax)))) {
            fprintf(stderr,"memory allocation error\n");
            sbs->n=0;
            return;
        }
    }
    sbs->msgs[sbs->n++]=sbsp;
}
/* add ion/utc parameter -------------------------------------------------------
* add ion/utc parameter to navigation messages
* args   : nav_t  *nav   IO     navigation messages
*          nav_t  *navp  I      ephemeris records
*-----------------------------------------------------------------------------*/
extern void addionutc(nav_t *nav, nav_t *navp)
{
    int i;
    
    tracet(4,"addionutc:\n");
    
    if (!nav) return;
    for (i=0;i<8;i++) nav->ion[i]=navp->ion[i];
    for (i=0;i<4;i++) nav->utc[i]=navp->utc[i];
    nav->leaps=navp->leaps;
}
/* convert receiver log file ---------------------------------------------------
* convert receiver log file to rinex obs/nav, sbas log files
* args   : char   *file  I      receiver log file
*          int     type  I      receiver log file type (RCVLOG_???)
*          char   *ofile I      output rinex obs file ("": no output)
*          char   *nfile I      output rinex nav file ("": no output)
*          char   *sfile I      output sbas log file ("": no output)
*          gtime_t ts,te I      start time/end time
*          double  tint  I      time interval (s)
*          rnxopt_t *rnxopt I   rinex options
*          int     opt   I      convert options
* return : status (1:0k,0:error)
*-----------------------------------------------------------------------------*/
extern int convlog(const char *file, int type, const char *ofile,
                   const char *nfile, const char *sfile, gtime_t ts, gtime_t te,
                   double tint, const rnxopt_t *rnxopt, int opt)
{
    FILE *fp,*ofp=NULL,*nfp=NULL,*sfp=NULL;
    
    tracet(3,"convlog: file=%s type=%d\n",file,type);
    
    if (!(fp=fopen(file,"rb"))) {
        return 0;
    }
    if (*ofile&&!(ofp=fopen(ofile,"w"))) {
        fclose(fp);
        return 0;
    }
    if (*nfile&&!(nfp=fopen(nfile,"w"))) {
        fclose(fp); fclose(ofp);
        return 0;
    }
    if (*sfile&&!(sfp=fopen(sfile,"w"))) {
        fclose(fp); fclose(ofp); fclose(nfp);
        return 0;
    }
    switch (type) {
    case RCVLOG_NOV  : convnov(fp,ofp,nfp,sfp,ts,te,tint,rnxopt,opt);  break;
    case RCVLOG_UBX  : convubx(fp,ofp,nfp,sfp,ts,te,tint,rnxopt,opt);  break;
    case RCVLOG_SS2  : convss2(fp,ofp,nfp,sfp,ts,te,tint,rnxopt,opt);  break;
    case RCVLOG_CRES : convcres(fp,ofp,nfp,sfp,ts,te,tint,rnxopt,opt); break;
    case RCVLOG_GARM : convgarm(fp,ofp,nfp,sfp,ts,te,tint,rnxopt,opt); break;
    case RCVLOG_SIRF : break;
    }
    fclose(fp); fclose(ofp); fclose(nfp); fclose(sfp);
    return 1;
}
/* read receiver log file and add records ------------------------------------*/
static int readlogf(const char *file, int type, gtime_t ts, gtime_t te,
                    double tint, int opt, obs_t *obs, nav_t *nav, sbs_t *sbs,
                    int *nmax)
{
    FILE *fp;
    
    tracet(3,"readlogf: file=%s type=%d\n",file,type);
    
    if (type==RCVLOG_RINEX) {
        return readrnxf(file,ts,te,tint,0,obs,nav,nmax,nmax+1);
    }
    else if (type==RCVLOG_SBAS) {
        if (sbs) sbs->n=sbsreadmsgt(file,0,ts,te,&sbs->msgs);
        return 1;
    }
    if (!(fp=fopen(file,"rb"))) {
        fprintf(stderr,"log file open error : %s\n",file);
        return 0;
    }
    switch (type) {
    case RCVLOG_NOV  : readnov(fp,ts,te,tint,opt,obs,nav,sbs,nmax);  break;
    case RCVLOG_UBX  : readubx(fp,ts,te,tint,opt,obs,nav,sbs,nmax);  break;
    case RCVLOG_SS2  : readss2(fp,ts,te,tint,opt,obs,nav,sbs,nmax);  break;
    case RCVLOG_CRES : readcres(fp,ts,te,tint,opt,obs,nav,sbs,nmax); break;
    case RCVLOG_GARM : readgarm(fp,ts,te,tint,opt,obs,nmax); break;
    case RCVLOG_SIRF : break;
    }
    fclose(fp);
    return 1;
}
/* read log file ---------------------------------------------------------------
* read receiver log file
* args   : char   *file  I      receiver log file
*          int     type  I      receiver log file type (RCVLOG_???)
*          gtime_t ts,te I      start time/end time
*          double  tint  I      time interval (s)
*          int     opt   I      convert options
*          obs_t  *obs   O      observation data (NULL: no output)
*          nav_t  *nav   O      naviagation messages (NULL: no output)
*          sbs_t  *sbs   O      sbas messages (NULL: no output)
* return : number of observation data
*-----------------------------------------------------------------------------*/
extern int readlog(const char *file, int type, gtime_t ts, gtime_t te,
                   double tint, int opt, obs_t *obs, nav_t *nav, sbs_t *sbs)
{
    int nmax[3]={0};
    
    tracet(3,"readlog : file=%s type=%d\n",file,type);
    
    if (obs) {obs->n=0; obs->data=NULL;}
    if (nav) {nav->n=0; nav->eph =NULL;}
    if (sbs) {sbs->n=0; sbs->msgs=NULL;}
    return readlogf(file,type,ts,te,tint,opt,obs,nav,sbs,nmax);
}
/* read receiver log files --------------------------------------------------------------
* read receiver log files
* args   : char  **files I      receiver log files
*          int    *type  I      receiver log file types (RCVLOG_???)
*          int    n      I      number of receiver log files
*          gtime_t ts,te I      start time/end time
*          double  tint  I      time interval (s)
*          int     opt   I      convert options
*          obs_t  *obs   O      observation data (NULL: no output)
*          nav_t  *nav   O      naviagation messages (NULL: no output)
*          sbs_t  *sbs   O      sbas messages (NULL: no output)
* return : number of observation data
*-----------------------------------------------------------------------------*/
extern int readlogs(char **files, int *types, int n, gtime_t ts, gtime_t te,
                    double tint, int opt, obs_t *obs, nav_t *nav, sbs_t *sbs)
{
    int i,j,m,nobs=0,rcv=1,nmax[3]={0};
    
    tracet(3,"readlogs: n=%d\n",n);
    
    if (obs) {obs->n=0; obs->data=NULL;}
    if (nav) {nav->n=0; nav->eph =NULL;}
    if (sbs) {sbs->n=0; sbs->msgs=NULL;}
    for (i=0;i<n;i++) {
        m=obs->n;
        if (!readlogf(files[i],types[i],ts,te,tint,opt,obs,nav,sbs,nmax)) continue;
        if (obs->n>=m) {
            for (j=m;j<obs->n;j++) obs->data[j].rcv=rcv;
            rcv++;
        }
    }
    if (obs) {
        if (!(obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*obs->n)))
            obs->n=0;
        else nobs=sortobs(obs);
    }
    if (nav) {
        if (!(nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*nav->n)))
            nav->n=0;
        else uniqeph(nav);
    }
    if (sbs) {
        if (!(sbs->msgs=(sbsmsg_t *)realloc(sbs->msgs,sizeof(sbsmsg_t)*sbs->n)))
           sbs->n=0;
    }
    return nobs;
}
/* sync message header -------------------------------------------------------*/
#define SYNCNOV(p)  (*(p)==0xAA&&*((p)+1)==0x44&&*((p)+2)==0x12)
#define SYNCUBX(p)  (*(p)==0xB5&&*((p)+1)==0x62)
#define SYNCSS2(p)  (*(p)==0x01&&(*((p)+1)^*((p)+2))==0xFF)
#define SYNCCRES(p) (*(p)=='$'&&*((p)+1)=='B'&&*((p)+2)=='I'&&*((p)+3)=='N')
#define SYNCNMEA(p) (*(p)=='$'&&*((p)+1)=='G'&&*((p)+2)=='P')
#define SYNCNMET(p) (*(p)==0x0D&&*((p)+1)==0x0A)

/* search message head -------------------------------------------------------*/
static unsigned char *searchhead(unsigned char *p, int n, int type)
{
    unsigned char *q;
    
    if (type==RCVLOG_NOV) {
        for (q=p;q<p+n-3;q++) if (SYNCNOV(q)) return q;
    }
    else if (type==RCVLOG_UBX) {
        for (q=p;q<p+n-2;q++) if (SYNCUBX(q)) return q;
    }
    else if (type==RCVLOG_SS2) {
        for (q=p;q<p+n-2;q++) if (SYNCSS2(q)) return q;
    }
    else if (type==RCVLOG_CRES) {
        for (q=p;q<p+n-4;q++) if (SYNCCRES(q)) return q;
    }
    else if (type==RCVLOG_NMEA) {
        for (q=p;q<p+n-2;q++) if (SYNCNMEA(q)) return q;
    }
    return NULL;
}
/* search message tail -------------------------------------------------------*/
static unsigned char *searchtail(unsigned char *p, int len, int type)
{
    unsigned char *q;
    int siz;
    
    if (type==RCVLOG_NOV) {
        if (len<10) return NULL;
        siz=*((unsigned short *)(p+8))+32;
        return len<siz?NULL:p+siz;
    }
    else if (type==RCVLOG_UBX) {
        if (len<6) return NULL;
        siz=*((unsigned short *)(p+4))+8;
        return len<siz?NULL:p+siz;
    }
    else if (type==RCVLOG_SS2) {
        if (len<4) return NULL;
        siz=*(p+3)+6;
        return len<siz?NULL:p+siz;
    }
    else if (type==RCVLOG_CRES) {
        if (len<8) return NULL;
        siz=*((unsigned short *)(p+6))+12;
        return len<siz?NULL:p+siz;
    }
    else if (type==RCVLOG_NMEA) {
        for (q=p;q<p+len-1&&q<p+256;q++) if (SYNCNMET(q)) return q+2;
    }
    return NULL;
}
/* decode receiver log ---------------------------------------------------------
* decode receiver log and extract a message
* args   : raw_t  *raw   IO     raw data decode control struct
*          unsigned char *buff I message buffer
*          int    len    I      buffer length (bytes)
*          int    type   I      receiver log file type (RCVLOG_???)
*          int    opt    I      decode option
*          obs_t  *obs   IO     observation data (NULL: no output)
*          nav_t  *nav   IO     navigation message (NULL: no output)
*          sbsmsg_t *sbsmsg IO  sbas message (NULL: no output)
*          int    *stat  O      status (-1:error, 0:no valid data, 1:obs data,
*                                       2:ephemeris, 3:sbas message,
*                                       9:ion/utc parameters)
* return : pointer to search next message
*          (NULL: no message, buff: imcomplete message)
*-----------------------------------------------------------------------------*/
extern unsigned char *decodelog(raw_t *raw, unsigned char *buff, int len,
                                int type, int opt, obs_t *obs, nav_t *nav,
                                sbsmsg_t *sbsmsg, int *stat)
{
    unsigned char *p,*q;
    int sat;
    
    *stat=0;
    
    tracet(4,"decodelog: len=%d type=%d\n",len,type);
    
    if (len<4) return buff; /* incomplete */
    
    if (!(p=searchhead(buff,len,type))) return NULL;
    
    if (!(q=searchtail(p,len-(p-buff),type))) { /* incomplete */
        return p;
    }
    if (type==RCVLOG_NOV) {
        *stat=decodenov(raw,p,q-p,opt,obs,nav,sbsmsg,&sat);
        tracet(5,"decodelog: decodenov len=%d stat=%d nobs=%d sat=%d\n",q-p,*stat,obs->n,sat);
        return q;
    }
    else if (type==RCVLOG_UBX) {
        *stat=decodeubx(raw,p+2,q-p-2,opt,obs,nav,sbsmsg,&sat);
        tracet(5,"decodelog: decodeubx len=%d stat=%d nobs=%d sat=%d\n",q-p,*stat,obs->n,sat);
        return q;
    }
    else if (type==RCVLOG_SS2) {
        *stat=decodess2(raw,p,q-p,opt,obs,nav,sbsmsg,&sat);
        tracet(5,"decodelog: decodess2 len=%d stat=%d nobs=%d sat=%d\n",q-p,*stat,obs->n,sat);
        return q;
    }
    else if (type==RCVLOG_CRES) {
        *stat=decodecres(raw,p,q-p,opt,obs,nav,sbsmsg,&sat);
        tracet(5,"decodelog: decodecres len=%d stat=%d nobs=%d sat=%d\n",q-p,*stat,obs->n,sat);
        return q;
    }
    return NULL;
}
