/*------------------------------------------------------------------------------
* rtkpos.c : realtime kinematic positioning
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2007/01/12 1.0 new
*           2007/03/13 1.1 add slip detection by LLI flag
*           2007/04/18 1.2 add antenna pcv correction
*                          change rtkpos argin
*           2008/07/18 1.3 refactored
*           2009/01/02 1.4 modify rtk positioning api
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: rtkpos.c,v 1.1 2008/07/17 21:48:06 ttaka Exp $";

/* constants/macros ----------------------------------------------------------*/

#define SQR(x)      ((x)*(x))
#define II(s,opt)   ((opt->mode!=5?3:0)+(s)-1)     /* state index of ionos */
#define IT(r,opt)   (II(1,opt)+(opt->estiono?MAXSAT:0)+(r)) /* state index of tropos */
#define NR(opt)     (IT(0,opt)+(opt->esttrop?2:0)) /* number of real prms */
#define IB(s,f,opt) (NR(opt)+MAXSAT*(f)+(s)-1)     /* state index of bias */

const static double fion[]={    /* ionospheric delay factor relative to P1 */
    1.0,FREQ1*FREQ1/FREQ2/FREQ2,FREQ1*FREQ1/FREQ5/FREQ5
};
/* single-differenced observable (f=0:L1,1:L2,2:C1,3:P2) ---------------------*/
static double sdobs(const obsd_t *obs, int i, int j, int f)
{
    double pi=f<NFREQ?lam[f]*obs[i].L[f]:obs[i].P[f-NFREQ];
    double pj=f<NFREQ?lam[f]*obs[j].L[f]:obs[j].P[f-NFREQ];
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced geometry-free linear combination of phase --------------*/
static double gfobs(const obsd_t *obs, int i, int j)
{
    double pi=sdobs(obs,i,j,0),pj=sdobs(obs,i,j,1);
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced measurement error variance (f=0:L1,1:L2,2:C1,3:P2) -----*/
static double varerr(double el, double bl, double dt, int f, const prcopt_t *opt)
{
    double a=opt->err[1],b=opt->err[2]/sin(el),c=opt->err[3]*bl/1E4;
    double d=CLIGHT*opt->sclkstab*dt;
    return 2.0*(a*a+b*b+c*c)*(f<opt->nf?1.0:opt->err[0]*opt->err[0])+d*d;
}
/* baseline length -----------------------------------------------------------*/
static double baseline(const double *ru, const double *rb)
{
    double dr[3];
    dr[0]=ru[0]-rb[0]; dr[1]=ru[1]-rb[1]; dr[2]=ru[2]-rb[2];
    return norm(dr,3);
}
/* initialize state and covariance -------------------------------------------*/
static void initx(rtk_t *rtk, double xi, double var, int i)
{
    int j;
    rtk->x[i]=xi;
    for (j=0;j<rtk->nx;j++) {
        rtk->P[i+j*rtk->nx]=rtk->P[j+i*rtk->nx]=i==j?var:0.0;
    }
}
/* select common satellites between rover and reference station --------------*/
static int selsat(const obsd_t *obs, double *azel, int nu, int nr,
                  const prcopt_t *opt, int *sat, int *iu, int *ir)
{
    int i,j,k=0;
    
    trace(3,"selsat : nu=%d nr=%d\n",nu,nr);
    
    for (i=0,j=nu;i<nu&&j<nu+nr;i++,j++) {
        if      (obs[i].sat<obs[j].sat) j--;
        else if (obs[i].sat>obs[j].sat) i--;
        else if (azel[1+j*2]>=opt->elmin&& /* elevation at base station */
                 (obs[i].SNR[0]==0||opt->snrmin<=obs[i].SNR[0])&&
                 (obs[j].SNR[0]==0||opt->snrmin<=obs[j].SNR[0])) {
            sat[k]=SAT(obs[i].sat); iu[k]=i; ir[k++]=j;
            trace(4,"(%2d) PRN%3d iu=%2d ir=%2d\n",k-1,obs[i].sat,i,j);
        }
    }
    return k;
}
/* temporal update of states -------------------------------------------------*/
static void udstate(rtk_t *rtk, const obsd_t *obs, const int *sat,
                    const int *iu, const int *ir, int ns)
{
    prcopt_t *opt=&rtk->opt;
    double tt,g0,g1,cp,pr,bias[MAXSAT]={0},offset,bl=baseline(rtk->x,rtk->rr[1]);
    int i,j,k,f,reset;
    
    trace(3,"udstate: ns=%d\n",ns);
    
    tt=fabs(rtk->tt);
    
    for (i=0;i<ns&&opt->mode>PMODE_DGPS;i++) {
        for (f=0;f<opt->nf;f++) { /* detect slip by LLI */
            rtk->slip[sat[i]-1][f]=obs[iu[i]].LLI[f]|obs[ir[i]].LLI[f];
            if (f>=1) rtk->slip[sat[i]-1][f]&=~1; /* clear L2 LLI flag */
            if (rtk->slip[sat[i]-1][f]) {
                trace(2,"udstate: slip detected (PRN%2d L%d LLI=%x %x)\n",
                      PRN(sat[i]),f+1,obs[iu[i]].LLI[f],obs[ir[i]].LLI[f]);
            }
        }
        if (opt->nf>=2&&(g1=gfobs(obs,iu[i],ir[i]))!=0.0) {
            g0=rtk->gf[sat[i]-1]; rtk->gf[sat[i]-1]=g1;
            if (g0!=0.0&&fabs(g1-g0)>opt->thresslip) { /* detect slip by GF jump */
                trace(2,"udstate: slip detected (PRN%2d GF=%.3f %.3f)\n",
                      PRN(sat[i]),g0,g1);
                for (f=0;f<opt->nf;f++) rtk->slip[sat[i]-1][f]|=1;
            }
        }
    }
    for (f=0;f<opt->nf&&opt->mode>PMODE_DGPS;f++) {
        /* reset phase-bias if instantaneous AR or expire obs outage counter */
        for (i=1;i<=MAXSAT;i++) {
            reset=++rtk->outc[i-1][f]>(unsigned int)opt->maxout;
            if (opt->modear==1&&rtk->x[IB(i,f,opt)]!=0.0) {
                initx(rtk,0.0,0.0,IB(i,f,opt));
            }
            else if (reset&&rtk->x[IB(i,f,opt)]!=0.0) {
                initx(rtk,0.0,0.0,IB(i,f,opt));
                trace(2,"udstate: obs outage counter overflow (PRN%2d L%d n=%d)\n",
                      PRN(i),f+1,rtk->outc[i-1][f]);
            }
            if (opt->modear!=1&&reset) {
                rtk->lock[i-1][f]=-opt->minlock;
            }
        }
        for (i=0;i<ns;i++) { /* reset phase-bias if slip */
            k=IB(sat[i],f,opt); rtk->P[k+k*rtk->nx]+=opt->prn[0]*opt->prn[0]*tt;
            if (opt->modear==1||!(rtk->slip[sat[i]-1][f]&1)) continue;
            rtk->x[k]=0.0; rtk->lock[sat[i]-1][f]=-opt->minlock;
        }
        for (i=j=0,offset=0.0;i<ns;i++) { /* estimate approx phase-bias */
            cp=sdobs(obs,iu[i],ir[i],f); pr=sdobs(obs,iu[i],ir[i],NFREQ);
            if (cp!=0.0&&pr!=0.0) bias[i]=(cp-pr)/lam[f]; 
            if (rtk->x[IB(sat[i],f,opt)]!=0.0) {
                offset+=bias[i]-rtk->x[IB(sat[i],f,opt)];
                j++;
            }
        }
        if (j>0) { /* assure phase-code coherency */
            for (i=1;i<=MAXSAT;i++) {
                if (rtk->x[IB(i,f,opt)]!=0.0) rtk->x[IB(i,f,opt)]+=offset/j;
            }
        }
        for (i=0;i<ns;i++) { /* set initial phase-bias */
            if (bias[i]==0.0||rtk->x[IB(sat[i],f,opt)]!=0.0) continue;
            initx(rtk,bias[i],SQR(opt->std[0]),IB(sat[i],f,opt));
        }
    }
    if (opt->estiono) { /* reset/update ionosphere parameter */
        for (i=1;i<=MAXSAT;i++) {
            if (rtk->x[II(i,opt)]!=0.0&&rtk->outc[i-1][0]>60&&rtk->outc[i-1][1]>60)
                rtk->x[II(i,opt)]=0.0;
        }
        for (i=0;i<ns;i++) {
            k=II(sat[i],opt);
            if (rtk->x[k]==0.0) initx(rtk,1E-6,SQR(opt->std[1]*bl/1E4),k);
            else rtk->P[k+k*rtk->nx]+=SQR(opt->prn[1]*bl/1E4)*tt;
        }
    }
    if (opt->esttrop) { /* reset/update tropospheric parameter */
        for (i=0;i<2;i++) {
            k=IT(i,opt);
            if (rtk->x[k]==0.0) initx(rtk,1E-6,SQR(opt->std[2]),k);
            else rtk->P[k+k*rtk->nx]+=SQR(opt->prn[2])*tt;
        }
    }
}
/* undifferenced phase/code residuals ----------------------------------------*/
static void zdres(const obsd_t *obs, int n, const double *rs, const double *dts,
                  const double *rr, const prcopt_t *opt, const pcv_t *pcv,
                  double *y, double *e, double *azel)
{
    double r,pos[3],dr[NFREQ]={0};
    int i,j;
    
    trace(3,"zdres  : n=%d\n",n);
    
    for (i=0;i<n*opt->nf*2;i++) y[i]=0.0;
    ecef2pos(rr,pos);
    
    for (i=0;i<n;i++) {
        if (SAT(obs[i].sat)>MAXSAT) continue;
        if ((r=geodist(rs+i*3,rr,e+i*3))<=0.0) continue;
        if (satazel(pos,e+i*3,azel+i*2)<opt->elmin) continue;
        
        /* satellite clock-bias and troposphere delay model */
        r+=-CLIGHT*dts[i]+tropmodel(pos,azel+i*2,0.5);
        
        /* antenna phase center correction */
        if (pcv) antmodel(pcv,azel+i*2,dr);
        
        for (j=0;j<opt->nf;j++) { /* observable - pseudorange */
            if (obs[i].L[j]!=0.0) y[j+i*opt->nf*2]=lam[j]*obs[i].L[j]-r-dr[j];
            if (obs[i].P[j]!=0.0) y[j+i*opt->nf*2+opt->nf]=obs[i].P[j]-r-dr[j];
        }
    }
    for (i=0;i<n;i++) {
        trace(4,"PRN%3d %13.3f %13.3f %13.3f %13.10f %6.1f %5.1f\n",
              obs[i].sat,rs[i*3],rs[1+i*3],rs[2+i*3],dts[i],azel[i*2]*R2D,
              azel[1+i*2]*R2D);
    }
    trace(4,"y=\n"); tracemat(4,y,opt->nf*2,n,13,3);
}
/* double-differenced phase/code residuals -----------------------------------*/
static int ddres(rtk_t *rtk, double dt, double *x, const int *sat, double *y,
                 double *e, double *azel, const int *iu, const int *ir, int ns,
                 double *v, double *H, double *R)
{
    prcopt_t *opt=&rtk->opt;
    double bl,posu[3],posr[3],didxi=0.0,didxj=0.0,*im,*tmu,*tmr,*Ri,*Rj,*Hi,s;
    int i,j,k,f,nv=0,ny=ns*opt->nf*2,n[NFREQ*2]={0};
    
    trace(3,"ddres  : dt=%.1f nx=%d ns=%d\n",dt,rtk->nx,ns);
    
    im=mat(ns,1); tmu=mat(ns,1); tmr=mat(ns,1); Ri=mat(ny,1); Rj=mat(ny,1);
    
    bl=baseline(x,rtk->rr[1]); ecef2pos(x,posu); ecef2pos(rtk->rr[1],posr);
    
    for (i=0;i<MAXSAT;i++) for (j=0;j<NFREQ;j++) rtk->res[i][j]=0.0;
    
    for (i=0;i<ns;i++) {
        if (opt->estiono) {
            im[i]=(ionmapf(posu,azel+iu[i]*2)+ionmapf(posr,azel+ir[i]*2))/2.0;
        }
        if (opt->esttrop) {
            tropmapf(rtk->time[0],posu,azel+iu[i]*2,tmu+i);
            tropmapf(rtk->time[0],posr,azel+ir[i]*2,tmr+i);
        }
    }
    for (f=opt->mode>PMODE_DGPS?0:opt->nf;f<opt->nf*2;f++) {
        
        /* make double-difference */
        for (i=j=0;j<ns;j++) {
            if (y[f+iu[j]*opt->nf*2]==0.0||y[f+ir[j]*opt->nf*2]==0.0) continue;
            if (azel[1+iu[j]*2]>=azel[1+iu[i]*2]) i=j; /* highest elevation */
        }
        for (j=0;j<ns;j++) {
            if (i==j) continue;
            if (y[f+iu[j]*opt->nf*2]==0.0||y[f+ir[j]*opt->nf*2]==0.0) continue;
              
            /* double-differenced residual */
            v[nv]=(y[f+iu[i]*opt->nf*2]-y[f+ir[i]*opt->nf*2])-
                  (y[f+iu[j]*opt->nf*2]-y[f+ir[j]*opt->nf*2]);
            
            if (opt->estiono) {
                didxi=(f<opt->nf?-1.0:1.0)*fion[f%opt->nf]*im[i];
                didxj=(f<opt->nf?-1.0:1.0)*fion[f%opt->nf]*im[j];
                v[nv]-=didxi*x[II(sat[i],opt)]-didxj*x[II(sat[j],opt)];
            }
            if (opt->esttrop) {
                v[nv]-=(tmu[i]-tmu[j])*x[IT(0,opt)]-(tmr[i]-tmr[j])*x[IT(1,opt)];
            }
            if (f<opt->nf) {
                v[nv]-=lam[f]*(x[IB(sat[i],f,opt)]-x[IB(sat[j],f,opt)]);
            }
            if (fabs(v[nv])>opt->maxinno) {
                trace(1,"ddres  : outlier rejected (PRN%2d-%2d %s%d v=%.3f)\n",
                      PRN(sat[i]),PRN(sat[j]),f<opt->nf?"L":"P",f%opt->nf+1,v[nv]);
                continue;
            }
            if (H) { /* partial derivatives */
                Hi=H+nv*rtk->nx; for (k=0;k<3;k++) Hi[k]=-e[k+iu[i]*3]+e[k+iu[j]*3];
                if (opt->estiono) {
                    Hi[II(sat[i],opt)]= didxi;
                    Hi[II(sat[j],opt)]=-didxj;
                }
                if (opt->esttrop) {
                    Hi[IT(0,opt)]= (tmu[i]-tmu[j]);
                    Hi[IT(1,opt)]=-(tmr[i]-tmr[j]);
                }
                if (f<opt->nf) {
                    Hi[IB(sat[i],f,opt)]= lam[f];
                    Hi[IB(sat[j],f,opt)]=-lam[f];
                }
            }
            if (opt->mode>PMODE_DGPS) {
                if (f<opt->nf) rtk->vsat[sat[i]-1][f]=rtk->vsat[sat[j]-1][f]=1;
            }
            else {
                rtk->vsat[sat[i]-1][f-opt->nf]=rtk->vsat[sat[j]-1][f-opt->nf]=1;
            }
            Ri[nv]=varerr(azel[1+iu[i]*2],bl,dt,f,opt);
            Rj[nv]=varerr(azel[1+iu[j]*2],bl,dt,f,opt);
            
            if (f<opt->nf) rtk->res[sat[j]-1][f]=v[nv];
            
            trace(4,"PRN%3d-%3d %s%d v=%13.3f R=%8.6f %8.6f\n",PRN(sat[i]),
                  PRN(sat[j]),f<opt->nf?"L":"P",f%opt->nf+1,v[nv],Ri[nv],Rj[nv]);
            nv++; n[f]++;
        }
        if (f<opt->nf) { /* single-differenced residuals */
            for (j=0,s=0.0;j<MAXSAT;j++) s+=rtk->res[j][f];
            s/=n[f]+1;
            for (j=0;j<MAXSAT;j++) {
                if (j==sat[i]-1||rtk->res[j][f]!=0.0) rtk->res[j][f]-=s;
            }
        }
    }
    if (H) {trace(5,"H=\n"); tracemat(5,H,rtk->nx,nv,7,4);}
    
    /* measurement error covariance */
    for (i=0;i<nv*nv;i++) R[i]=0.0;
    for (f=k=0;f<opt->nf*2;k+=n[f++]) {
        for (i=k;i<k+n[f];i++) {
            for (j=i;j<k+n[f];j++) R[i+j*nv]=R[j+i*nv]=Ri[i];
            R[i+i*nv]+=Rj[i];
        }
    }
    trace(5,"R=\n"); tracemat(5,R,nv,nv,8,6);
    
    free(im); free(tmu); free(tmr); free(Ri); free(Rj);
    return nv;
}
/* time-interpolation of residuals -------------------------------------------*/
static double intpres(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                      const double *rb, const prcopt_t *opt, double *y)
{
    static obsd_t obsb[MAXOBS];
    static double yb[MAXOBS*NFREQ*2];
    static double rs[MAXOBS*3],dts[MAXOBS],e[MAXOBS*3],azel[MAXOBS*2];
    static int nb=0;
    double dt=timediff(time,obs[0].time),dtb,*p,*q;
    int i,j,k;
    
    trace(3,"intpres: n=%d dt=%.1f\n",n,dt);
    
    if (nb==0||fabs(dt)<DTTOL) {
        nb=n; for (i=0;i<n;i++) obsb[i]=obs[i];
        return dt;
    }
    dtb=timediff(time,obsb[0].time);
    if (fabs(dtb)>opt->maxtdiff*2.0||dtb==dt) return dt;
    
    if (opt->sateph) {
        satposp(time,obsb,nb,nav,rs,dts);
    }
    else {
        satpos(time,obsb,nb,nav,rs,dts);
    }
    zdres(obsb,nb,rs,dts,rb,opt,opt->pcvr,yb,e,azel);
    for (i=0;i<n;i++) {
        for (j=0;j<nb;j++) if (obsb[j].sat==obs[i].sat) break;
        if (j>=nb) continue;
        for (k=0,p=y+i*opt->nf*2,q=yb+j*opt->nf*2;k<opt->nf*2;k++,p++,q++) {
            if (*p==0.0||*q==0.0) *p=0.0; else *p=(dtb*(*p)-dt*(*q))/(dtb-dt);
        }
    }
    return fabs(dtb)>fabs(dt)?dtb:dt;
}
/* relative positioning ------------------------------------------------------*/
static int relpos(rtk_t *rtk, const obsd_t *obs, int nu, int nr,
                  const nav_t *nav)
{
    prcopt_t *opt=&rtk->opt;
    gtime_t time=obs[0].time;
    double *rs,*dts,*y,*e,*azel,*v,*H,*R,*xp,*Pp,dt,vv;
    int i,j,f,n=nu+nr,ns,ny,nv,sat[MAXSAT],iu[MAXSAT],ir[MAXSAT],info=-1;
    
    trace(3,"relpos : nx=%d nu=%d nr=%d\n",rtk->nx,nu,nr);
    
    dt=timediff(time,obs[nu].time);
    
    rs=mat(3,n); dts=mat(1,n); y=mat(opt->nf*2,n); e=mat(3,n); azel=zeros(2,n);
    
    for (i=0;i<MAXSAT;i++) for (j=0;j<NFREQ;j++) {
        rtk->vsat[i][j]=rtk->slip[i][j]=0;
    }
    /* satellite position/clock-bias */
    if (rtk->opt.sateph) {
        satposp(time,obs,n,nav,rs,dts);
    }
    else {
        satpos(time,obs,n,nav,rs,dts);
    }
    /* residuals of base station */
    zdres(obs+nu,nr,rs+nu*3,dts+nu,rtk->rr[1],opt,opt->pcvr,y+nu*opt->nf*2,
          e+nu*3,azel+nu*2);
    
    /* time-interpolation of residuals */
    if (opt->intpref) {
        dt=intpres(time,obs+nu,nr,nav,rtk->rr[1],opt,y+nu*opt->nf*2);
    }
    /* select common satellites */
    if ((ns=selsat(obs,azel,nu,nr,opt,sat,iu,ir))<=0) {
        trace(1,"relpos : no common satellite\n");
        free(rs); free(dts); free(y); free(e); free(azel);
        return -1;
    }
    /* temporal update of states */
    udstate(rtk,obs,sat,iu,ir,ns);
    
    trace(4,"x(0)="); tracemat(4,rtk->x,1,NR(opt),13,4);
    
    ny=ns*opt->nf*2;
    xp=mat(rtk->nx,1); Pp=mat(rtk->nx,rtk->nx);
    v=mat(ny,1); H=zeros(rtk->nx,ny); R=mat(ny,ny);
    matcpy(xp,rtk->x,rtk->nx,1);
    
    for (i=0;i<opt->niter;i++) {
        matcpy(Pp,rtk->P,rtk->nx,rtk->nx);
        
        /* rover undifferenced residuals */
        zdres(obs,nu,rs,dts,xp,opt,opt->pcvu,y,e,azel);
        
        /* double-differenced residuals */
        if ((nv=ddres(rtk,dt,xp,sat,y,e,azel,iu,ir,ns,v,H,R))<1) {
            trace(1,"relpos : no double-differenced residual\n");
            info=-1; break;
        }
        /* kalman filter measurement update */
        if ((info=filter(xp,Pp,H,v,R,rtk->nx,nv))) {
            trace(1,"relpos : filter error (info=%d)\n",info);
            break;
        }
        trace(4,"x(%d)=",i+1); tracemat(4,xp,1,NR(opt),13,4);
    }
    if (!info) { /* validation by postfit-residual test */
        zdres(obs,nu,rs,dts,xp,opt,opt->pcvu,y,e,azel);
        nv=ddres(rtk,dt,xp,sat,y,e,azel,iu,ir,ns,v,NULL,R);
        if (nv>NR(opt)) {
            for (i=0,vv=0.0;i<nv;i++) vv+=v[i]*v[i]/R[i+i*nv];
            if (vv>chisqr[nv-NR(opt)-1]) info=-9;
            trace(!info?2:1,"relpos : validation %s (nv=%d nr=%d vv=%.2f cs=%.2f)\n",
                  !info?"ok":"error",nv,NR(opt),vv,chisqr[nv-NR(opt)-1]);
        }
    }
    if (!info) { /* update state, covariance and ambiguity control struct */
        matcpy(rtk->x,xp,rtk->nx,1);
        matcpy(rtk->P,Pp,rtk->nx,rtk->nx);
        rtk->ns=0;
        for (i=0;i<ns;i++) for (f=0;f<opt->nf;f++) {
            if (!rtk->vsat[sat[i]-1][f]) continue;
            rtk->lock[sat[i]-1][f]++;
            rtk->outc[sat[i]-1][f]=0;
            if (azel[1+iu[i]*2]<opt->elmaskar) rtk->lock[sat[i]-1][f]=0;
            if (f==0) rtk->ns++;
        }
    }
    
    free(rs); free(dts); free(y); free(e); free(azel);
    free(xp); free(Pp); free(v); free(H); free(R);
    return info;
}
/* resolve integer ambiguity -------------------------------------------------*/
static int resamb(rtk_t *rtk)
{
    prcopt_t *opt=&rtk->opt;
    int i,j,k,f,ny,nb=0,info,*indi,*indj,nx=rtk->nx,na=rtk->na;
    double *D,*DP,*y,*Qy,*b,*db,*Qb,*Qab,*QQ,s[2];
    
    trace(3,"resamb : nx=%d\n",nx);
    
    D=zeros(nx,nx); indi=imat(MAXSAT*opt->nf,1); indj=imat(MAXSAT*opt->nf,1);
    
    /* single to double-difference transformation matrix (D') */
    for (i=0;i<na;i++) D[i+i*nx]=1.0;
    for (f=0,k=na;f<opt->nf;f++,k+=MAXSAT) {
        for (i=k;i<k+MAXSAT;i++) {
            if (rtk->x[i]!=0.0&&rtk->lock[i-k][f]>0&&!(rtk->slip[i-k][f]&2)) break;
        }
        for (j=i+1;j<k+MAXSAT;j++) {
            if (rtk->x[j]!=0.0&&rtk->lock[j-k][f]>0&&!(rtk->slip[j-k][f]&2)) {
                D[i+(na+nb)*nx]=1.0; D[j+(na+nb)*nx]=-1.0;
                indi[nb]=i; indj[nb++]=j;
            }
        }
    }
    if (nb<=0) {
        trace(1,"resamb : no valid double-difference\n");
        free(D); free(indi); free(indj);
        return -1;
    }
    ny=na+nb; y=mat(ny,1); Qy=mat(ny,ny); DP=mat(ny,nx);
    b=mat(nb,2); db=mat(nb,1); Qb=mat(nb,nb); Qab=mat(na,nb); QQ=mat(na,nb);
    
    trace(5,"D=\n"); tracemat(5,D,nx,ny,2,0);
    
    /* transform single to double-differenced phase-bias (y=D'*x, Qy=D'*P*D) */
    matmul("TN",ny,1,nx,1.0,D,rtk->x,0.0,y);
    matmul("TN",ny,nx,nx,1.0,D,rtk->P,0.0,DP);
    matmul("NN",ny,ny,nx,1.0,DP,D,0.0,Qy);
    
    /* phase-bias covariance (Qb) and real-parameters to bias covariance (Qab) */
    for (i=0;i<nb;i++) for (j=0;j<nb;j++) Qb[i+j*nb]=Qy[na+i+(na+j)*ny];
    for (i=0;i<na;i++) for (j=0;j<nb;j++) Qab[i+j*na]=Qy[i+(na+j)*ny];
    
    trace(4,"N(0)="); tracemat(4,y+na,1,nb,10,3);
    
    /* lambda/mlambda integer least-square estimation */
    if (!(info=lambda(nb,2,y+na,Qb,b,s))) {
        
        trace(4,"N(1)="); tracemat(4,b,1,nb,10,3);
        trace(4,"N(2)="); tracemat(4,b+nb,1,nb,10,3);
        
        rtk->ambratio=s[0]>0?s[1]/s[0]:0.0;
        
        /* validation by popular ratio-test */
        if (s[0]<=0.0||s[1]/s[0]>=opt->thresar) {
            
            /* transform float to fixed solution (xa=xa-Qab*Qb\(b0-b)) */
            for (i=0;i<na;i++) {
                rtk->xa[i]=rtk->x[i];
                for (j=0;j<na;j++) rtk->Pa[i+j*na]=rtk->P[i+j*nx];
            }
            for (i=0;i<nb;i++) y[na+i]-=b[i];
            if (!(info=matinv(Qb,nb))) {
                matmul("NN",nb,1,nb,1.0,Qb,y+na,0.0,db);
                matmul("NN",na,1,nb,-1.0,Qab,db,1.0,rtk->xa);
                
                /* covariance (Qa=Qa-Qab*Qb^-1*Qab') */
                matmul("NN",na,nb,nb,1.0,Qab,Qb,0.0,QQ);
                matmul("NT",na,na,nb,-1.0,QQ,Qab,1.0,rtk->Pa);
            }
        }
        else info=-1; /* validation failed */
        
        trace(!info?2:1,"resamb : validation %s (nb=%d ratio=%.2f s=%.2f/%.2f)\n",
              !info?"ok":"error",nb,s[1]/s[0],s[0],s[1]);
    }
    else trace(1,"resamb : lambda error (info=%d)\n",info);
    
    free(D); free(indi); free(indj); free(y); free(Qy); free(DP);
    free(b); free(db); free(Qb); free(Qab); free(QQ);
    return info; /* 0:ok */
}
/* initialize rtk control ------------------------------------------------------
* initialize rtk control struct
* args   : rtk_t    *rtk    IO  rtk control/result struct
*          prcopt_t *opt    I   positioning options (see rtklib.h)
* return : none
*-----------------------------------------------------------------------------*/
extern void rtkinit(rtk_t *rtk, const prcopt_t *opt)
{
    rtk_t rtk0={{{0}}};
    *rtk=rtk0;
    rtk->nx=NR(opt)+MAXSAT*opt->nf;
    rtk->na=NR(opt);
    rtk->x=zeros(rtk->nx,1);
    rtk->P=zeros(rtk->nx,rtk->nx);
    rtk->xa=zeros(rtk->na,1);
    rtk->Pa=zeros(rtk->na,rtk->na);
    rtk->opt=*opt;
}
/* free rtk control ------------------------------------------------------------
* free memory for rtk control struct
* args   : rtk_t    *rtk    IO  rtk control/result struct
* return : none
*-----------------------------------------------------------------------------*/
extern void rtkfree(rtk_t *rtk)
{
    free(rtk->x); rtk->x=NULL;
    free(rtk->P); rtk->P=NULL;
    free(rtk->xa); rtk->xa=NULL;
    free(rtk->Pa); rtk->Pa=NULL;
}
/* estimate receiver position/velocity by single point positining ------------*/
static int rcvposv(rtk_t *rtk, const obsd_t *obs, int nobs, const nav_t *nav,
                   const sbsstat_t *sbsstat, int index, double *Qr, double *Qv)
{
    gtime_t time;
    double dtr[2],azel[MAXOBS*2];
    double elmin=rtk->opt.elmin,snrmin=rtk->opt.snrmin;
    int i,corr,vsat[MAXOBS]={0},stat=SOLQ_SINGLE;
    
    trace(3,"rcvposv : nobs=%d nnav=%d\n",nobs,nav->n);
    
    /* receiver position */
    if (sbsstat&&rtk->opt.sbascorr) {
        rtk->ns=sbspntpos(obs,nobs,nav,sbsstat,elmin,snrmin,rtk->opt.sbascorr,
                          rtk->rr[index],Qr,dtr,azel,vsat,&corr);
        if (corr) stat=SOLQ_SBAS;
    }
    else {
        rtk->ns=pntpos(obs,nobs,nav,elmin,snrmin,rtk->rr[index],Qr,dtr,azel,vsat);
    }
    /* azimuth/elevation and valid satellite flags */
    for (i=0;i<MAXSAT;i++) rtk->vs[index][i]=0;
    for (i=0;i<nobs;i++) {
        rtk->az[index][SAT(obs[i].sat)-1]=azel[i*2];
        rtk->el[index][SAT(obs[i].sat)-1]=azel[1+i*2];
        rtk->vs[index][SAT(obs[i].sat)-1]=vsat[i];
    }
    if (rtk->ns<4) return SOLQ_NONE;
    
    /* receiver time */
    time=timeadd(obs[0].time,-dtr[0]);
    if (index==0&&rtk->time[0].time!=0) rtk->tt=timediff(time,rtk->time[0]);
    rtk->time[index]=time;
    
    /* receiver velocity */
    if (pntvel(obs,nobs,nav,rtk->rr[index],azel,vsat,rtk->rr[index]+3,Qv,
               dtr+1)<4) {
        for (i=0;i<3;i++) rtk->rr[index][i+3]=0.0;
    }
    return stat;
}
/* rtk positioning -------------------------------------------------------------
* input observation data and navigation message, compute rover position by 
* relative positioning and output results.
* args   : rtk_t *rtk       IO  rtk control/result struct
*          obsd_t *obs      I   observation data for an epoch
*                               obs->data[i].rcv=1:rover,2:reference
*                               sorted by receiver and satellte
*          int    nobs      I   number of observation data
*          nav_t  *nav      I   navigation messages
*          sbsstat_t *sbsstat I sbas status (NULL: no correction)
* return : solution status (0:error,1:fix,2:float,4:dgps,5:single)
*-----------------------------------------------------------------------------*/
extern int rtkpos(rtk_t *rtk, const obsd_t *obs, int nobs, const nav_t *nav,
                  const sbsstat_t *sbsstat)
{
    prcopt_t *opt=&rtk->opt;
    double tt,Qr[2][9]={{0}},Qv[2][9]={{0}};
    int i,j,nu,nr,stat;
    char tstr[64];
    
    time2str(obs[0].time,tstr,3);
    trace(2,"rtkpos : time=%s nobs=%d nnav=%d\n",tstr,nobs,nav->n);
    trace(4,"obs=\n"); traceobs(4,obs,nobs);
    trace(4,"nav=\n"); tracenav(4,nav);
    
    if (opt->mode!=PMODE_MOVEB) {
        for (i=0;i<3;i++) rtk->rr[1][i]=opt->rb[i];
    }
    /* count rover/base station observations */
    for (nu=0;nu<nobs&&obs[nu].rcv==1;nu++) ;
    for (nr=0;nu+nr<nobs&&obs[nu+nr].rcv==2;nr++) ;
    if (opt->mode>PMODE_SINGLE) {
        if (nu==0||nr==0) {
            trace(1,"rtkpos : number of obs error (nu=%d nr=%d)\n",nu,nr);
            return SOLQ_NONE;
        }
        rtk->age=timediff(obs[0].time,obs[nu].time);
        if (fabs(rtk->age)>opt->maxtdiff) {
            trace(1,"rtkpos : age of differential error (age=%.1f)\n",rtk->age);
            return SOLQ_NONE;
        }
    }
    /* rover position/velocity */
    if ((stat=rcvposv(rtk,obs,nu,nav,sbsstat,0,Qr[0],Qv[0]))==SOLQ_NONE) {
        trace(1,"rtkpos : initial rover position error\n");
        return SOLQ_NONE;
    }
    if (opt->mode==PMODE_SINGLE) {
        for (i=0;i<3;i++) {
            rtk->x[i]=rtk->rr[0][i];
            for (j=i;j<rtk->nx;j++) {
                rtk->P[i+j*rtk->nx]=rtk->P[j+i*rtk->nx]=j<3?Qr[0][i+j*3]:0.0;
            }
        }
        return stat;
    }
    if (opt->mode==PMODE_MOVEB) {
        /* base station position/velocity */
        if (rcvposv(rtk,obs+nu,nr,nav,sbsstat,1,Qr[1],Qv[1])==SOLQ_NONE) {
            trace(1,"rtkpos : base station position error\n");
            return SOLQ_NONE;
        }
        /* receiver time difference correction */
        tt=timediff(rtk->time[0],rtk->time[1]);
        for (i=0;i<3;i++) rtk->rr[1][i]+=rtk->rr[1][i+3]*tt;
    }
    /* first epoch or kinematic mode */
    if (norm(rtk->x,3)<=0.0||
        opt->mode==PMODE_DGPS||opt->mode==PMODE_KINEMA||opt->mode==PMODE_MOVEB) {
        for (i=0;i<3;i++) {
            rtk->x[i]=rtk->rr[0][i];
            for (j=i;j<rtk->nx;j++) {
                rtk->P[i+j*rtk->nx]=rtk->P[j+i*rtk->nx]=j<3?Qr[0][i+j*3]:0.0;
            }
        }
    }
    /* rover position by relative positioning */
    if (relpos(rtk,obs,nu,nr,nav)) {
        return SOLQ_NONE;
    }
    stat=opt->mode<=PMODE_DGPS?SOLQ_DGPS:SOLQ_FLOAT;
    
    /* resolve integer ambiguity */
    if (opt->mode>=2&&opt->thresar>=1.0&&!resamb(rtk)) stat=SOLQ_FIX;
    
    return stat;
}
